/*
 This file is part of the digitalSTROM ds485-client library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-client library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM ds485-client library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM ds485-client library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <malloc.h>
#include <syslog.h>
#include <pthread.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <digitalSTROM/ds485-core.h>
#include <digitalSTROM/ds485.h>
#include "ds485-client-interface.h"
#include "ds485-client.h"

/**
 * Local Data
 */
static ds485_client_interface_t connInterfaceSerial =
{
        .conntype = CLIENT_SERIAL,
        .open = ds485_serial_client_open,
        .close = ds485_serial_client_close,
        .send = ds485_serial_client_send_command,
        .get_bus_devices = ds485_serial_client_query_devices,
        .get_bus_state = ds485_serial_client_query_state,
        .get_dsid = ds485_serial_client_get_dsid,
};

static void* serial_thread(void* arg);
static void serial_cleanup(void *arg);
static int serial_callback_state(dS485Device *dev, bus_state_t state);
static int serial_callback_bus_change(dS485Device *dev, dsid_t *id, int flags);
static int serial_callback_system_msg(dS485Device *dev, unsigned char level, const char *msg, va_list args);

/**
 * Initialize serial client data
 */
int ds485_serial_client_init(ds485_client_t *cl)
{
    if (!strncmp(cl->connspec, "serial://", 9)) {
        return ERROR_INVALID_CONNSPEC;
    }

    char *connection = strdup(cl->connspec + 9);
    char *part = strtok(connection, ":");
    if (part == NULL) {
        return ERROR_INVALID_CONNSPEC;
    }
    free(connection);

    cl->serialClient.dName = strdup(part);
    cl->serialClient.rxfilterFlags = DS485_FILTER_STANDARD;
    cl->connInterface = connInterfaceSerial;
    return 0;
}

int ds485_serial_client_open(ds485_client_t *cl, const dsid_t *ownDSID)
{
    int loopback = 1;

    cl->serialClient.ownDSID = *ownDSID;
    cl->serialClient.serialHandle = ds485_api_open(cl->serialClient.dName, ownDSID);
    if (cl->serialClient.serialHandle == 0) {
        pthread_mutex_lock(&cl->openQuery.mutex);
        cl->openQuery.state = STATE_RECEIVED;
        pthread_cond_signal(&cl->openQuery.condition);
        pthread_mutex_unlock(&cl->openQuery.mutex);
        return -1;
    }

    /* Start receiver thread */
    if (pthread_create(&cl->connThread, NULL, serial_thread, cl)) {
    }

    /* enable loopback mode */
    ds485_api_ioctl(cl->serialClient.serialHandle, IOC_LOOPBACK_MODE, &loopback, sizeof(loopback));

    /* register message handler */
    ds485_api_ioctl(cl->serialClient.serialHandle, IOC_SET_MSG_CB, serial_callback_system_msg, sizeof(serial_callback_system_msg));

    ds485_api_ioctl(cl->serialClient.serialHandle, IOC_SET_BUS_STATE_CB, serial_callback_state, sizeof(serial_callback_state));

    ds485_api_ioctl(cl->serialClient.serialHandle, IOC_SET_BUS_CHANGE_CB, serial_callback_bus_change, sizeof(serial_callback_bus_change));

    int logger = open("/tmp/ds485", O_WRONLY | O_NONBLOCK);
    if (logger < 0)
    {
        printf("Could not open logger file.\n");
    }
    else
    {
        ds485_api_ioctl(cl->serialClient.serialHandle, IOC_SET_LOGGER_FD, &logger, sizeof(logger));
    }

    pthread_mutex_lock(&cl->openQuery.mutex);
    cl->openQuery.state = STATE_RECEIVED;
    pthread_cond_signal(&cl->openQuery.condition);
    pthread_mutex_unlock(&cl->openQuery.mutex);

    return 0;
}

/**
 * Closes the session, frees all connection related resources
 */
int ds485_serial_client_close(ds485_client_t *cl)
{
    cl->active = 0;

    /* cleanup rx thread */
    pthread_cancel(cl->connThread);

    free(cl->serialClient.dName);

    return ds485_api_close(cl->serialClient.serialHandle);
}

/**
 * Query the list of devices on the dS485 bus
 */
int ds485_serial_client_query_devices(ds485_client_t *cl, dsid_t *device_list, int device_list_size)
{
    cl->busQuery.length = ds485_api_get_bus_members(cl->serialClient.serialHandle, device_list, device_list_size);
    cl->busQuery.state = STATE_RECEIVED;
    pthread_cond_signal(&cl->busQuery.condition);

    return cl->busQuery.length;
}

/**
 * Query the current bus state
 */
int ds485_serial_client_query_state(ds485_client_t *cl, bus_state_t *state)
{
    cl->busQuery.length = ds485_api_get_bus_state(cl->serialClient.serialHandle);

    cl->busQuery.state = STATE_RECEIVED;
    pthread_cond_signal(&cl->busQuery.condition);

    return 0;
}

int ds485_serial_client_get_dsid(ds485_client_t *cl)
{
    cl->busQuery.devices[0] = cl->serialClient.ownDSID;
    cl->busQuery.state = STATE_RECEIVED;
    pthread_cond_signal(&cl->busQuery.condition);

    return 0;
}

/**
 * Send dM-API command
 */
int ds485_serial_client_send_command(ds485_client_t *cl, ds485_container_t *request)
{
    request->transactionId = 10;

    if (ds485_api_send_data(cl->serialClient.serialHandle, request)) {
        printf("serial client: send failed\n");
        return -1;
    }

    return 0;
}

/*************************** Protocol Callbacks *******************************/

/**
 * registered system callback for protocol (debug) messages
 */
static int serial_callback_system_msg(dS485Device *dev, unsigned char level, const char *msg, va_list args)
{
    char buffer[512];

    vsnprintf(buffer, 512, msg, args);

    //if (verbose > level)
    printf("%s", buffer);

    return 0;
}

/**
 * registered device callback for bus state changes
 */
static int serial_callback_state(dS485Device *dev, bus_state_t state)
{
    if (NULL == dev)
        return -1;

    switch (state)
    {
    case DS485_ISOLATED:
        printf("STATE: ISOLATED\n");
        break;
    case DS485_CONNECTED:
        printf("STATE: CONNECTED\n");
        break;
    case DS485_ACTIVE:
        printf("STATE: ACTIVE\n");
        break;
    case DS485_JOIN:
        printf("STATE: JOIN\n");
        break;
    default:
        printf("STATE: *UNKNOWN*\n");
        break;
    }
    return 0;
}

static int serial_callback_bus_change(dS485Device *dev, dsid_t *id, int flags)
{
    return -1;
}

static void serial_cleanup(void *arg)
{
    ds485_client_t *cl = (ds485_client_t *) arg;

    free(cl->serialClient.dName);
    ds485_api_close(cl->serialClient.serialHandle);

    ds485_cleanup(cl);
}

/******************************* RECEIVE THREAD *******************************/

static void* serial_thread(void* arg)
{
    ds485_client_t *cl = (ds485_client_t *) arg;
    ds485_container_t container;
    int rval;

    pthread_cleanup_push(&serial_cleanup, cl);

    printf("Receive thread running.\n");
    while (cl->active) {

        rval = ds485_api_recv_data(cl->serialClient.serialHandle, &container);

        if (rval < 0) {
            continue;
        }

        pthread_mutex_lock(&cl->syncSend.mutex);
        if (cl->syncSend.state == STATE_WAITING) {
            if ((IsEqualId(cl->syncSend.response.sourceId, container.sourceId)) &&
                (cl->syncSend.response.containerType == container.containerType) &&
                (cl->syncSend.response.data[0] == container.data[0]))
            {
                cl->syncSend.response = container;
                cl->syncSend.state = STATE_RECEIVED;
                pthread_cond_signal(&cl->syncSend.condition);
                pthread_mutex_unlock(&cl->syncSend.mutex);
                continue;
            }
        }
        pthread_mutex_unlock(&cl->syncSend.mutex);

        if (cl->connInterface.responseCallback) {
            cl->connInterface.responseCallback(cl->connInterface.responseCallbackData, &container);
        }
    }
    pthread_cleanup_pop(1);

    printf("Receive thread terminated.\n");
    return NULL;
}
